@@ -1,3 +1,28 @@
+#' Run \code{system2()} and mark its character output as UTF-8 if appropriate
+#'
+#' This is a wrapper function based on \code{system2()}. If \code{system2()}
+#' returns character output (e.g., with the argument \code{stdout = TRUE}),
+#' check if the output is encoded in UTF-8. If it is, mark it with UTF-8
+#' explicitly.
+#' @param ... Passed to \code{\link{system2}()}.
+#' @return The value returned by \code{system2()}.
+#' @export
+#' @examplesIf interactive()
+#' a = shQuote(c('-e', 'print(intToUtf8(c(20320, 22909)))'))
+#' x2 = system2('Rscript', a, stdout = TRUE)
+#' Encoding(x2)  # unknown
+#'
+#' x3 = xfun::system3('Rscript', a, stdout = TRUE)
+#' # encoding of x3 should be UTF-8 if the current locale is UTF-8
+#' !l10n_info()[['UTF-8']] || Encoding(x3) == 'UTF-8'  # should be TRUE
+system3 = function(...) {
+  res = system2(...)
+  if (is.character(res)) {
+    if (all(is_utf8(res))) Encoding(res) = 'UTF-8'
+  }
+  if (is.integer(res) && res == 0) invisible(res) else res
+}
+
 #' Run OptiPNG on all PNG files under a directory
 #'
 #' Call the command \command{optipng} via \code{system2()} to optimize all PNG
@@ -72,7 +97,7 @@ Rscript_call = function(
     c(options, shQuote(c(pkg_file('scripts', 'call-fun.R'), f)))
     ,..., wait = wait
   )
-  if (wait) if (file.exists(f[2])) readRDS(f[2]) else stop(fail, call. = FALSE)
+  if (wait) if (file_exists(f[2])) readRDS(f[2]) else stop(fail, call. = FALSE)
 }
 
 # call a function in a background process
@@ -126,7 +151,7 @@ proc_kill = function(pid, recursive = TRUE, ...) {
 
 # obtain pids of all child processes (recursively)
 child_pids = function(id) {
-  x = system2('sh', shQuote(c(pkg_file('scripts', 'child_pids.sh'), id)), stdout = TRUE)
+  x = system2('sh', shQuote(c(pkg_file('scripts', 'child-pids.sh'), id)), stdout = TRUE)
   grep('^[0-9]+$', x, value = TRUE)
 }
 
@@ -211,7 +236,7 @@ bg_process = function(
     get_pid = function() {
       # make sure the command points to an actual executable (e.g., resolve 'R'
       # to 'R.exe')
-      if (!file.exists(command)) {
+      if (!file_exists(command)) {
         if (Sys.which(command) != '') command = Sys.which(command)
       }
       cmd = basename(command)
@@ -272,11 +297,13 @@ upload_ftp = function(file, server, dir = '') {
   }
 }
 
+#' @param solaris Whether to also upload the package to the Rhub server to check
+#'   it on Solaris.
 #' @rdname upload_ftp
 #' @export
 upload_win_builder = function(
   file = pkg_build(), version = c("R-devel", "R-release", "R-oldrelease"),
-  server = c('ftp', 'https')
+  server = c('ftp', 'https'), solaris = pkg_available('rhub')
 ) {
   if (missing(file)) on.exit(file.remove(file), add = TRUE)
   if (system2('git', 'status', stderr = FALSE) == 0) system2('git', 'pull')
@@ -316,5 +343,11 @@ upload_win_builder = function(
       }
     })
   }
+
+  if (solaris) rhub::check_on_solaris(
+    file, check_args = '--no-manual', show_status = FALSE,
+    env_vars = c(`_R_CHECK_FORCE_SUGGESTS_` = 'false')
+  )
+
   setNames(unlist(res), version)
 }
