@@ -78,6 +78,25 @@ same_path = function(p1, p2, ...) {
   normalize_path(p1, ...) == normalize_path(p2, ...)
 }
 
+#' Find file paths that exist
+#'
+#' This is a shorthand of \code{x[file.exists(x)]}, and optionally returns the
+#' first existing file path.
+#' @param x A vector of file paths.
+#' @param first Whether to return the first existing path. If \code{TRUE} and no
+#'   specified files exist, it will signal an error.
+#' @return A vector of existing file paths.
+#' @export
+#' @examples
+#' xfun::existing_files(c('foo.txt', system.file('DESCRIPTION', package='xfun')))
+existing_files = function(x, first = FALSE) {
+  x = x[file_exists(x)]
+  if (!first) return(x)
+  x = head(x, 1)
+  if (length(x) != 1) stop('None of the specified files exist.')
+  x
+}
+
 #' Return the (possible) root directory of a project
 #'
 #' Given a path of a file (or dir) in a potential project (e.g., an R package or
@@ -111,7 +130,7 @@ proj_root = function(path = './', rules = root_rules) {
   dir = if (dir_exists(path)) path else dirname(path)
   if (same_path(dir, file.path(dir, '..'))) return()
   if (is.null(dim(rules))) dim(rules) = c(1, length(rules))
-  for (i in seq_along(nrow(rules))) {
+  for (i in seq_len(nrow(rules))) {
     file = rules[i, 1]; pattern = rules[i, 2]
     for (f in list.files(dir, file, full.names = TRUE)) {
       if (pattern == '' || length(grep(pattern, read_utf8(f)))) return(dir)
@@ -127,27 +146,34 @@ root_rules = matrix(c(
   '.+[.]Rproj$',   '^Version: '
 ), ncol = 2, byrow = TRUE, dimnames = list(NULL, c('file', 'pattern')))
 
-#' Get the relative path of a path relative a directory
+#' Get the relative path of a path relative to a directory
 #'
 #' Given a directory, return the relative path that is relative to this
 #' directory. For example, the path \file{foo/bar.txt} relative to the directory
 #' \file{foo/} is \file{bar.txt}, and the path \file{/a/b/c.txt} relative to
 #' \file{/d/e/} is \file{../../a/b/c.txt}.
 #' @param dir Path to a directory.
-#' @param x The path to be converted to a relative path.
+#' @param x A vector of paths to be converted to relative paths.
 #' @param use.. Whether to use double-dots (\file{..}) in the relative path. A
 #'   double-dot indicates the parent directory (starting from the directory
 #'   provided by the \code{dir} argument).
-#' @param error Whether to signal an error if the path cannot be converted to a
+#' @param error Whether to signal an error if a path cannot be converted to a
 #'   relative path.
-#' @return A relative path if the conversion succeeded; otherwise the original
-#'   path when \code{error = FALSE}, and an error when \code{error = TRUE}.
+#' @return A vector of relative paths if the conversion succeeded; otherwise the
+#'   original paths when \code{error = FALSE}, and an error when \code{error =
+#'   TRUE}.
 #' @export
 #' @examples
 #' xfun::relative_path('foo/bar.txt', 'foo/')
 #' xfun::relative_path('foo/bar/a.txt', 'foo/haha/')
 #' xfun::relative_path(getwd())
 relative_path = function(x, dir = '.', use.. = TRUE, error = TRUE) {
+  res = x
+  for (i in seq_along(x)) res[i] = relative_path_one(x[i], dir, use.., error)
+  res
+}
+
+relative_path_one = function(x, dir, use.., error) {
   # on Windows, if a relative path doesn't exist, normalizePath() will use
   # getwd() as its parent dir; however, normalizePath() just returns the
   # relative path on *nix, and we have to assume it's relative to getwd()
@@ -419,10 +445,12 @@ print.xfun_rename_seq = function(x, ...) {
   print(tab)
 }
 
-# return path to R's svg logo if it exists, otherwise return the jpg logo
-R_logo = function() {
+# return path to R's svg logo if it exists, otherwise return the jpg logo; or
+# specify a regex to match the logo path, e.g., ext = 'jpg$'
+R_logo = function(ext = NULL) {
   x = file.path(R.home('doc'), 'html', c('Rlogo.svg', 'logo.jpg'))
-  x[file.exists(x)][1]
+  if (!is.null(ext)) x = grep(ext, x, value = TRUE)
+  existing_files(x, first = TRUE)
 }
 
 #' Extract filenames from a URLs
